#ifndef __FT2_MIX_MACROS_H
#define __FT2_MIX_MACROS_H

#include "ft2_header.h"

/* ----------------------------------------------------------------------- */
/*                          GENERAL MIXER MACROS                           */
/* ----------------------------------------------------------------------- */

#define GET_MIXER_VARS \
    audioMixL  = audio.mixBufferL; \
    audioMixR  = audio.mixBufferR; \
    mixInMono  = (v->SLVol2 == v->SRVol2); \
    realPos    = v->SPos; \
    pos        = v->SPosDec; \
    delta      = v->SFrq; \

#define GET_MIXER_VARS_RAMP \
    audioMixL  = audio.mixBufferL; \
    audioMixR  = audio.mixBufferR; \
    CDA_LVolIP = v->SLVolIP; \
    CDA_RVolIP = v->SRVolIP; \
    mixInMono  = (v->SLVol2 == v->SRVol2) && (CDA_LVolIP == CDA_RVolIP); \
    realPos    = v->SPos; \
    pos        = v->SPosDec; \
    delta      = v->SFrq; \

#define SET_BACK_MIXER_VARS \
    pos &= 0xFFFF; \
    v->SLVol2 = CDA_LVol; \
    v->SRVol2 = CDA_RVol; \

#define SET_BACK_MIXER_POS \
    v->SPosDec = pos; \
    v->SPos    = realPos; \

#define GET_VOL \
    CDA_LVol = v->SLVol2; \
    CDA_RVol = v->SRVol2; \

#define GET_POS           smpPos = realPos + (pos >> 16);
#define GET_POS_BACKWARDS smpPos = realPos - (pos >> 16);

#define INC_POS pos += delta;
#define DEC_POS pos += delta;

#define INC_REAL_POS realPos += (pos >> 16);
#define DEC_REAL_POS realPos -= (pos >> 16);



/* ----------------------------------------------------------------------- */
/*                          SAMPLE RENDERING MACROS                        */
/* ----------------------------------------------------------------------- */

/* linear interpolation */
#define _LERP(s1, s2, f) /* s1,s2 = -32768..32767 | f = 0..65535 (frac) */ \
    s2  -= s1; \
    s2  *= (int32_t)((f) >> 1); \
    s2 >>= 16; \
    s2  += s2; \
    s1  += s2; \

/* all the 64-bit MULs here convert to fast logic even on most 32-bit CPUs */

#define RENDER_8BIT_SMP \
    sample = v->sampleData8[smpPos] << (28 - 8); \
    *audioMixL++ += (int32_t)(((int64_t)(sample) * CDA_LVol) >> 32); \
    *audioMixR++ += (int32_t)(((int64_t)(sample) * CDA_RVol) >> 32); \

#define RENDER_8BIT_SMP_LERP \
    sample  = v->sampleData8[smpPos    ] << 8; \
    sample2 = v->sampleData8[smpPos + 1] << 8; \
    _LERP(sample, sample2, pos & 0xFFFF) \
    sample <<= (28 - 16); \
    *audioMixL++ += (int32_t)(((int64_t)(sample) * CDA_LVol) >> 32); \
    *audioMixR++ += (int32_t)(((int64_t)(sample) * CDA_RVol) >> 32); \

#define RENDER_8BIT_SMP_LERP_BACKWARDS \
    sample  = v->sampleData8[smpPos    ] << 8; \
    sample2 = v->sampleData8[smpPos + 1] << 8; \
    _LERP(sample, sample2, 0xFFFF - (pos & 0xFFFF)) \
    sample <<= (28 - 16); \
    *audioMixL++ += (int32_t)(((int64_t)(sample) * CDA_LVol) >> 32); \
    *audioMixR++ += (int32_t)(((int64_t)(sample) * CDA_RVol) >> 32); \

#define RENDER_16BIT_SMP \
    sample = v->sampleData16[smpPos] << (28 - 16); \
    *audioMixL++ += (int32_t)(((int64_t)(sample) * CDA_LVol) >> 32); \
    *audioMixR++ += (int32_t)(((int64_t)(sample) * CDA_RVol) >> 32); \

#define RENDER_16BIT_SMP_LERP \
    sample  = v->sampleData16[smpPos    ]; \
    sample2 = v->sampleData16[smpPos + 1]; \
    _LERP(sample, sample2, pos & 0xFFFF) \
    sample <<= (28 - 16); \
    *audioMixL++ += (int32_t)(((int64_t)(sample) * CDA_LVol) >> 32); \
    *audioMixR++ += (int32_t)(((int64_t)(sample) * CDA_RVol) >> 32); \

#define RENDER_16BIT_SMP_LERP_BACKWARDS \
    sample  = v->sampleData16[smpPos    ]; \
    sample2 = v->sampleData16[smpPos + 1]; \
    _LERP(sample, sample2, 0xFFFF - (pos & 0xFFFF)) \
    sample <<= (28 - 16); \
    *audioMixL++ += (int32_t)(((int64_t)(sample) * CDA_LVol) >> 32); \
    *audioMixR++ += (int32_t)(((int64_t)(sample) * CDA_RVol) >> 32); \

#define RENDER_8BIT_SMP_MONO \
    sample = v->sampleData8[smpPos] << (28 - 8); \
    sample = (int32_t)(((int64_t)(sample) * CDA_LVol) >> 32); \
    \
    *audioMixL++ += sample; \
    *audioMixR++ += sample; \

#define RENDER_8BIT_SMP_MONO_LERP \
    sample  = v->sampleData8[smpPos    ] << 8; \
    sample2 = v->sampleData8[smpPos + 1] << 8; \
    _LERP(sample, sample2, pos & 0xFFFF) \
    sample <<= (28 - 16); \
    sample = (int32_t)(((int64_t)(sample) * CDA_LVol) >> 32); \
    \
    *audioMixL++ += sample; \
    *audioMixR++ += sample; \

#define RENDER_8BIT_SMP_MONO_LERP_BACKWARDS \
    sample  = v->sampleData8[smpPos    ] << 8; \
    sample2 = v->sampleData8[smpPos + 1] << 8; \
    _LERP(sample, sample2, 0xFFFF - (pos & 0xFFFF)) \
    sample <<= (28 - 16); \
    sample = (int32_t)(((int64_t)(sample) * CDA_LVol) >> 32); \
    \
    *audioMixL++ += sample; \
    *audioMixR++ += sample; \

#define RENDER_16BIT_SMP_MONO \
    sample = v->sampleData16[smpPos] << (28 - 16); \
    sample = (int32_t)(((int64_t)(sample) * CDA_LVol) >> 32); \
    \
    *audioMixL++ += sample; \
    *audioMixR++ += sample; \

#define RENDER_16BIT_SMP_MONO_LERP \
    sample  = v->sampleData16[smpPos    ]; \
    sample2 = v->sampleData16[smpPos + 1]; \
    _LERP(sample, sample2, pos & 0xFFFF) \
    sample <<= (28 - 16); \
    sample = (int32_t)(((int64_t)(sample) * CDA_LVol) >> 32); \
    \
    *audioMixL++ += sample; \
    *audioMixR++ += sample; \

#define RENDER_16BIT_SMP_MONO_LERP_BACKWARDS \
    sample  = v->sampleData16[smpPos    ]; \
    sample2 = v->sampleData16[smpPos + 1]; \
    _LERP(sample, sample2, 0xFFFF - (pos & 0xFFFF)) \
    sample <<= (28 - 16); \
    sample = (int32_t)(((int64_t)(sample) * CDA_LVol) >> 32); \
    \
    *audioMixL++ += sample; \
    *audioMixR++ += sample; \

#define VOLUME_RAMPING \
    CDA_LVol += CDA_LVolIP; \
    CDA_RVol += CDA_RVolIP; \



/* ----------------------------------------------------------------------- */
/*                      SAMPLES-TO-MIX LIMITING MACROS                     */
/* ----------------------------------------------------------------------- */

#define LIMIT_MIX_NUM \
    CDA_SmpEndFlag = true; \
    \
    i = (v->SLen - 1) - realPos; \
    if (v->SFrq > (i >> 16)) \
    { \
        if (i > 65535) /* won't fit in a 32-bit div */ \
        { \
            samplesToMix = ((uint32_t)(0xFFFFFFFF - pos) / v->SFrq) + 1; \
            CDA_SmpEndFlag = false; \
        } \
        else \
        { \
            samplesToMix = ((uint32_t)((i << 16) | (0xFFFF - pos)) / v->SFrq) + 1; \
        } \
    } \
    else \
    { \
        samplesToMix = 65535; \
    } \
    \
    if (samplesToMix > (uint32_t)(CDA_BytesLeft)) \
    { \
        samplesToMix = CDA_BytesLeft; \
        CDA_SmpEndFlag = false; \
    } \

#define LIMIT_MIX_NUM_BIDI_LOOP \
    CDA_SmpEndFlag = true; \
    \
    if (v->backwards) \
        i = realPos - v->SRepS; \
    else \
        i = (v->SLen - 1) - realPos; \
    \
    if (v->SFrq > (i >> 16)) \
    { \
        if (i > 65535) /* won't fit in a 32-bit div */ \
        { \
            samplesToMix = ((uint32_t)(0xFFFFFFFF - pos) / v->SFrq) + 1; \
            CDA_SmpEndFlag = false; \
        } \
        else \
        { \
            samplesToMix = ((uint32_t)((i << 16) | (0xFFFF - pos)) / v->SFrq) + 1; \
        } \
    } \
    else \
    { \
        samplesToMix = 65535; \
    } \
    \
    if (samplesToMix > (uint32_t)(CDA_BytesLeft)) \
    { \
        samplesToMix = CDA_BytesLeft; \
        CDA_SmpEndFlag = false; \
    } \

#define LIMIT_MIX_NUM_RAMP \
    if (v->SVolIPLen <= 0) \
    { \
        CDA_LVolIP = 0; \
        CDA_RVolIP = 0; \
        \
        if (v->isFadeOutVoice) \
        { \
            v->mixRoutine = NULL; /* fade out voice is done, shut it down */ \
            return; \
        } \
    } \
    else \
    { \
        if (samplesToMix > (uint32_t)(v->SVolIPLen)) \
        { \
            samplesToMix = v->SVolIPLen; \
            CDA_SmpEndFlag = false; \
        } \
        \
        v->SVolIPLen -= samplesToMix; \
    } \



/* ----------------------------------------------------------------------- */
/*                     SAMPLE END/LOOP WRAPPING MACROS                     */
/* ----------------------------------------------------------------------- */
/*   NOTE: all MY_ASSERT() macros do absolutely nothing in release mode!   */

#ifdef _DEBUG
#define BIDI_WRAP_DEBUG \
    if (v->backwards) \
    { \
        MY_ASSERT(realPos < v->SRepS) \
    } \
    else \
    { \
        MY_ASSERT(realPos >= v->SLen) \
    }
#else
#define BIDI_WRAP_DEBUG
#endif

#define HANDLE_SAMPLE_END \
    if (CDA_SmpEndFlag) \
    { \
        MY_ASSERT(realPos >= v->SLen) \
        \
        v->mixRoutine = NULL; \
        return; \
    } \

#define WRAP_LOOP \
    if (CDA_SmpEndFlag) \
    { \
        MY_ASSERT(realPos >= v->SLen) \
        \
        do \
        { \
            realPos -= v->SRepL; \
        } \
        while (realPos >= v->SLen); \
    } \

#define WRAP_BIDI_LOOP \
    if (CDA_SmpEndFlag) \
    { \
        BIDI_WRAP_DEBUG \
        \
        /* if backwards: turn backward underflow into forward overflow */ \
        if (v->backwards) \
            realPos = (v->SLen - 1) + (v->SRepS - realPos); \
        \
        do \
        { \
            realPos -= v->SRepL; \
            v->backwards ^= 1; \
        } \
        while (realPos >= v->SLen); \
        \
        /* if backwards: forwards position -> backwards position */ \
        if (v->backwards) \
            realPos = (v->SLen - 1) - (realPos - v->SRepS); \
    } \



/* ----------------------------------------------------------------------- */
/*                       VOLUME=0 OPTIMIZATION MACROS                      */
/* ----------------------------------------------------------------------- */

#define VOL0_OPTIMIZATION_NO_LOOP \
    /* this translates to fast logic even on most 32-bit CPUs */ \
    pos64 = ((uint64_t)(numSamples) << 16) * v->SFrq; \
    realPos = (int32_t)(pos64 >> 32); \
    pos = (uint32_t)(pos64 & 0xFFFFFFFF) >> 16; \
    \
    realPos += v->SPos; \
    if ((pos + v->SPosDec) > 0xFFFF) \
        realPos++; \
    pos = (pos + v->SPosDec) & 0xFFFF; \
    \
    if (realPos >= v->SLen) \
    { \
        v->mixRoutine = NULL; \
        return; \
    } \
    \
    SET_BACK_MIXER_POS \

#define VOL0_OPTIMIZATION_LOOP \
    /* this translates to fast logic even on most 32-bit CPUs */ \
    pos64 = ((uint64_t)(numSamples) << 16) * v->SFrq; \
    realPos = (int32_t)(pos64 >> 32); \
    pos = (uint32_t)(pos64 & 0xFFFFFFFF) >> 16; \
    \
    realPos += v->SPos; \
    if ((pos + v->SPosDec) > 0xFFFF) \
        realPos++; \
    pos = (pos + v->SPosDec) & 0xFFFF; \
    \
    while (realPos >= v->SLen) \
           realPos -= v->SRepL; \
    \
    SET_BACK_MIXER_POS \

#define VOL0_OPTIMIZATION_BIDI_LOOP \
    /* this translates to fast 32-bit logic on most 32-bit CPUs */ \
    pos64 = ((uint64_t)(numSamples) << 16) * v->SFrq; \
    realPos = (int32_t)(pos64 >> 32); \
    pos = (uint32_t)(pos64 & 0xFFFFFFFF) >> 16; \
    \
    /* if backwards: backwards position -> forwards position */ \
    if (v->backwards) \
        v->SPos = (v->SLen - 1) - (v->SPos - v->SRepS); \
    \
    realPos = v->SPos + realPos; \
    if ((pos + v->SPosDec) > 0xFFFF) \
        realPos++; \
    pos = (pos + v->SPosDec) & 0xFFFF; \
    \
    while (realPos >= v->SLen) \
    { \
        realPos -= v->SRepL; \
        v->backwards ^= 1; \
    } \
    \
    /* if backwards: forwards position -> backwards position */ \
    if (v->backwards) \
        realPos = (v->SLen - 1) - (realPos - v->SRepS); \
    \
    SET_BACK_MIXER_POS \

#endif

/* get it? */
