/* for finding memory leaks in debug mode with Visual Studio */
#if defined _DEBUG && defined _MSC_VER
#include <crtdbg.h>
#endif

#include <stdio.h>
#include "ft2_header.h"
#include "ft2_audio.h"
#include "ft2_gui.h"
#include "ft2_mouse.h"
#include "ft2_sample_ed.h"
#include "ft2_module_loader.h"

/*
** This XM saver is directly ported, so it should act identical to FT2
** except that uninitialized data is initialized here (bug in FT2).
*/

static SDL_Thread *thread;

static uint16_t packPatt(uint8_t *pattPtr, uint16_t numRows);

/* ft2_replayer.c */
extern const char modSig[16][5];
extern const uint16_t amigaPeriod[12 * 8];

int8_t saveXM(UNICHAR *filenameU)
{
    uint8_t *pattPtr;
    int16_t ap, ai, i, k, a;
    uint16_t b, c;
    size_t result;
    instrTyp *ins;
    songHeaderTyp h;
    patternHeaderTyp ph;
    instrHeaderTyp ih;
    sampleTyp *srcSmp;
    sampleHeaderTyp *dstSmp;
    FILE *f;

    f = UNICHAR_FOPEN(filenameU, "wb");
    if (f == NULL)
    {
        setMouseBusy(false);
        sysReqQueue(SR_SAVE_IO_ERROR);
        return (false);
    }

    memcpy(h.sig, "Extended Module: ", 17);
    memset(h.name, ' ', 20);
    h.name[20] = 0x1A;
    memcpy(h.name, song.name, strlen(song.name));
    memcpy(h.progName, PROG_NAME_STR, 20);
    h.ver = 0x0104;
    h.headerSize = 20 + 256;
    h.len = song.len;
    h.repS = song.repS;
    h.antChn = song.antChn;
    h.defTempo = song.tempo;
    h.defSpeed = song.speed;

    /* count number of patterns */
    ap = MAX_PATTERNS;
    do
    {
        if (patternEmpty(ap - 1))
            ap--;
        else
            break;
    }
    while (ap > 0);
    h.antPtn = ap;

    /* count number of instruments */
    ai = MAX_INST;
    while ((ai > 0) && (getUsedSamples(ai) == 0) && (song.instrName[ai][0] == '\0'))
        ai--;
    h.antInstrs = ai;

    h.flags = linearFrqTab;
    memcpy(h.songTab, song.songTab, sizeof (song.songTab));

    if (fwrite(&h, sizeof (h), 1, f) != 1)
    {
        sysReqQueue(SR_SAVE_IO_ERROR);
        goto writeErrorXM;
    }

    for (i = 0; i < ap; ++i)
    {
        if (patternEmpty(i))
        {
            if (patt[i] != NULL)
            {
                free(patt[i]);
                patt[i] = NULL;
            }

            pattLens[i] = 64;
        }

        ph.patternHeaderSize = sizeof (patternHeaderTyp);
        ph.pattLen = pattLens[i];
        ph.typ = 0;

        if (patt[i] == NULL)
        {
            ph.dataLen = 0;
            if (fwrite(&ph, ph.patternHeaderSize, 1, f) != 1)
            {
                sysReqQueue(SR_SAVE_IO_ERROR);
                goto writeErrorXM;
            }
        }
        else
        {
            c = packPatt((uint8_t *)(patt[i]), pattLens[i]);
            b = pattLens[i] * TRACK_WIDTH;
            ph.dataLen = c;

            result  = fwrite(&ph,     ph.patternHeaderSize, 1, f);
            result += fwrite(patt[i], ph.dataLen,           1, f);

            pattPtr = (uint8_t *)(patt[i]);

            memcpy(&pattPtr[b - c], patt[i], c);
            unpackPatt(pattPtr, b - c, pattLens[i], song.antChn);
            clearUnusedChannels(patt[i], pattLens[i], song.antChn);

            if (result != 2) /* write was not OK */
            {
                sysReqQueue(SR_SAVE_IO_ERROR);
                goto writeErrorXM;
            }
        }
    }

    for (i = 1; i <= ai; ++i)
    {
        ins = &instr[i];

        a = getUsedSamples(i);

        memset(ih.name, 0, 22);
        memcpy(ih.name, song.instrName[i], strlen(song.instrName[i]));

        ih.typ        = 0;
        ih.antSamp    = a;
        ih.sampleSize = sizeof (sampleHeaderTyp);

        if (a > 0)
        {
            ih.instrSize = INSTR_HEADER_SIZE;

            memcpy(ih.ta, ins, INSTR_SIZE);
            for (k = 0; k < a; ++k)
            {
                srcSmp = &ins->samp[k];
                dstSmp = &ih.samp[k];

                memset(dstSmp->name, ' ', sizeof (dstSmp->name));

                memcpy(dstSmp, srcSmp, 12 + 4 + 2 + strlen(srcSmp->name));
                if (srcSmp->pek == NULL)
                    dstSmp->len = 0;
            }
        }
        else
        {
            ih.instrSize = 22 + 11;
        }

        if (fwrite(&ih, ih.instrSize + (a * sizeof (sampleHeaderTyp)), 1, f) != 1)
        {
            sysReqQueue(SR_SAVE_IO_ERROR);
            goto writeErrorXM;
        }

        for (k = 0; k < a; ++k)
        {
            srcSmp = &ins->samp[k];
            if (srcSmp->pek != NULL)
            {
                restoreSample(srcSmp);
                samp2Delta(srcSmp->pek, srcSmp->len, srcSmp->typ);

                result = fwrite(srcSmp->pek, srcSmp->len, 1, f);

                delta2Samp(srcSmp->pek, srcSmp->len, srcSmp->typ);
                fixSample(srcSmp);

                if (result != 1) /* write not OK */
                {
                    sysReqQueue(SR_SAVE_IO_ERROR);
                    goto writeErrorXM;
                }
            }
        }
    }

    removeSongModifiedFlag();

    fclose(f);

    editor.diskOpReadDir = true; /* force diskop re-read */

    setMouseBusy(false);
    return (true);

writeErrorXM:
    fclose(f);

    setMouseBusy(false);
    return (false);
}

/* this one is not ported from FT2, but who cares. XM saving accuracy is more important to me. */
static int8_t saveMOD(UNICHAR *filenameU)
{
    uint8_t orderOverflow, sampleOverflow, numPatt, tmpSmp, tmpTon, tmpEff, tmpEffParam;
    int16_t *ptr16, tmp16;
    uint16_t tmpPeriod, songLen;
    int32_t len, loopStart, loopLen, i, j, k;
    FILE *f;
    instrTyp *ins;
    sampleTyp *smp;
    tonTyp *t, emptyNote;

    orderOverflow  = false;
    sampleOverflow = false;

    songLen = song.len;
    if (songLen > 128)
    {
        songLen = 128;
        orderOverflow = true;
    }

    /* count number of patterns */
    numPatt = 0;
    for (i = 0; i < songLen; ++i)
    {
        if (song.songTab[i] >= 128)
        {
            setMouseBusy(false);
            sysReqQueue(SR_SAVE_ERR_PATT);
            return (false);
        }

        if (numPatt < song.songTab[i])
            numPatt = song.songTab[i];

        if (pattLens[song.songTab[i]] != 64)
        {
            setMouseBusy(false);
            sysReqQueue(SR_SAVE_ERR_PATTLEN);
            return (false);
        }
    }
    numPatt++;

    /* check if we used >31 instruments */
    for (i = 0; i < numPatt; ++i)
    {
        if (patt[i] == NULL)
            continue; /* pattern not allocated, skip it */

        for (j = 0; j < 64; ++j)
        {
            for (k = 0; k < song.antChn; ++k)
            {
                t = &patt[i][(j * MAX_VOICES) + k];
                if (t->instr > 31)
                {
                    setMouseBusy(false);
                    sysReqQueue(SR_SAVE_ERR_TOO_MANY_INSTR);
                    return (false);
                }
            }
        }
    }

    f = UNICHAR_FOPEN(filenameU, "wb");
    if (f == NULL)
    {
        setMouseBusy(false);
        sysReqQueue(SR_SAVE_IO_ERROR);
        return (false);
    }

    /* write song name */
    for (i = 0; i < 20; ++i)
    {
        if (i < (signed)(strlen(song.name)))
            fputc(song.name[i], f);
        else
            fputc('\0', f);
    }

    /* write samples */
    for (i = 0; i < 31; ++i)
    {
        ins = &instr[1 + i];
        smp = &ins->samp[0];

        len       = smp->len;
        loopStart = smp->repS;
        loopLen   = smp->repL;

        if (smp->typ & 16)
        {
            len       = (len       & 0xFFFFFFFE) / 2;
            loopStart = (loopStart & 0xFFFFFFFE) / 2;
            loopLen   = (loopLen   & 0xFFFFFFFE) / 2;
        }

        if (len > 131070)
        {
            len = 131070;
            sampleOverflow = true;
        }

        if ((smp->repS + smp->repL) <= len)
        {
            loopStart = smp->repS;
            loopLen   = smp->repL;
        }
        else
        {
            loopStart = 0;
            loopLen   = 2;
        }

        /* sample name */
        fwrite(song.instrName[1 + i], 1, 22, f);

        /* sample length */
        fputc(((len / 2) >> 8) & 0xFF, f);
        fputc( (len / 2)       & 0xFF, f);

        /* finetune and volume */
        fputc(((smp->fine + 128) >> 4) ^ 8, f);
        fputc((smp->vol > 64) ? 64 : smp->vol, f);

        /* loop start and loop length */
        fputc(((loopStart / 2) >> 8) & 0xFF, f);
        fputc( (loopStart / 2)       & 0xFF, f);
        fputc(((loopLen   / 2) >> 8) & 0xFF, f);
        fputc( (loopLen   / 2)       & 0xFF, f);
    }

    /* write song length and song repeat position */
    fputc(song.len & 0xFF, f);
    if (song.repS < song.len)
        fputc(song.repS & 0xFF, f);
    else
        fputc(0x7F, f);

    /* write order list */
    for (i = 0; i < 128; ++i)
    {
        if (i < songLen)
            fputc(song.songTab[i], f);
        else
            fputc(0, f);
    }

    /* write module type tag */
    if (song.antChn == 4)
    {
        if (numPatt <= 64)
            fwrite("M.K.", 1, 4, f);
        else
            fwrite("M!K!", 1, 4, f); /* FT2 doesn't do this... nasty */
    }
    else
    {
        if (fwrite(modSig[(song.antChn - 1) / 2], 4, 1, f) != 1)
        {
            sysReqQueue(SR_SAVE_IO_ERROR);
            goto writeErrorMOD;
        }
    }

    memset(&emptyNote, 0, sizeof (tonTyp));

    /* pattern data */
    for (i = 0; i < numPatt; ++i)
    {
        for (j = 0; j < 64; ++j)
        {
            for (k = 0; k < song.antChn; ++k)
            {
                if (patt[i] == NULL)
                    t = &emptyNote; /* pattern not allocated */
                else
                    t = &patt[i][(j * MAX_VOICES) + k];

                tmpSmp = t->instr; /* we know that instr can't be above 31 at this point */

                tmpTon = t->ton;
                if ((tmpTon >= 1) && (tmpTon <= (12 * 8)))
                    tmpPeriod = amigaPeriod[tmpTon - 1];
                else
                    tmpPeriod = 0;

                tmpEff      = t->effTyp;
                tmpEffParam = t->eff;

                if (t->effTyp == 20) /* Kxx */
                {
                    /* translate Kxx (key off) */
                    tmpEff      = 0xC;
                    tmpEffParam = 0x00;
                }
                else if (t->effTyp > 0x0F)
                {
                    tmpEff      = 0;
                    tmpEffParam = 0;
                }

                if ((tmpEff + tmpEffParam) == 0)
                {
                    /* we have a free effect slot, try to convert vol/keyoff effects.. */

                    if (t->ton == 97)
                    {
                        /* key off */
                        tmpEff = 0xC;
                        tmpEffParam = 0;
                    }
                    else
                    {
                        if ((t->vol >= 0x10) && (t->vol <= 0x50))
                        {
                            /* vol */
                            tmpEff = 0xC;
                            tmpEffParam = t->vol - 0x10;
                        }
                        else if ((t->vol >= 0xC0) && (t->vol <= 0xCF))
                        {
                            /* panning */
                            tmpEff = 0x8;
                            tmpEffParam = (t->vol - 0xC0) * 17;
                        }
                    }
                }

                /* write data */
                fputc((tmpSmp & 0xF0) | ((tmpPeriod >> 8) & 0x0F), f);
                fputc(tmpPeriod & 0x00FF, f);
                fputc(((tmpSmp << 4) & 0xF0) | (tmpEff & 0x0F), f);
                fputc(tmpEffParam, f);
            }
        }
    }

    /* write sample datas */
    for (i = 0; i < 31; ++i)
    {
        smp = &instr[1 + i].samp[0];

        len = smp->len;
        if (len > 0)
        {
            restoreSample(smp);

            if (smp->typ & 16)
            {
                /* 16-bit sample */
                ptr16 = (int16_t *)(smp->pek);

                len /= 2;
                if (len > 131070)
                    len = 131070;

                for (j = 0; j < len; ++j)
                {
                    tmp16 = ptr16[j];
                    fputc((int8_t)(tmp16 >> 8), f);
                }
            }
            else
            {
                /* 8-bit sample */

                if (len > 131070)
                    len = 131070;

                if (fwrite(smp->pek, len, 1, f) != 1)
                {
                    fixSample(smp);
                    sysReqQueue(SR_SAVE_IO_ERROR);
                    goto writeErrorMOD;
                }
            }

            fixSample(smp);
        }
    }

    fclose(f);
    removeSongModifiedFlag();

    if (orderOverflow)  sysReqQueue(SR_SAVE_WARN_ORDERS);
    if (sampleOverflow) sysReqQueue(SR_SAVE_WARN_SAMPLEN);
    if (linearFrqTab)   sysReqQueue(SR_SAVE_LINEAR_FREQ);

    editor.diskOpReadDir = true; /* force diskop re-read */

    setMouseBusy(false);
    return (true);

writeErrorMOD:
    fclose(f);

    setMouseBusy(false);
    return (false);
}

int32_t SDLCALL saveMusicThread(void *ptr)
{
    int8_t result;

    (void)(ptr);

    if (editor.tmpFilenameU == NULL)
    {
#ifdef _DEBUG
        __debugbreak();
#endif
        setMouseBusy(false);
        sysReqQueue(SR_SAVE_IO_ERROR);
        return (false);
    }

    lockMixerCallback();

    if (editor.moduleSaveMode == 1)
        result = saveXM(editor.tmpFilenameU);
    else
        result = saveMOD(editor.tmpFilenameU);

    unlockMixerCallback();

    return (result);
}

void saveMusic(UNICHAR *filenameU)
{
    UNICHAR_STRCPY(editor.tmpFilenameU, filenameU);

    setMouseBusy(true);
    thread = SDL_CreateThread(saveMusicThread, "FT2 Clone Module Saving Thread", NULL);
    if (thread == NULL)
    {
        setMouseBusy(false);
        sysReqQueue(SR_THREAD_ERROR);
        return;
    }

    /* don't let thread wait for this thread, let it clean up on its own when done */
    SDL_DetachThread(thread);
}

/* also directly ported from FT2 */
static uint16_t packPatt(uint8_t *pattPtr, uint16_t numRows)
{
    uint8_t bytes[sizeof (tonTyp)], packBits, *writePtr, *firstBytePtr;
    uint16_t row, chn, totalPackLen;

    totalPackLen = 0;

    if (pattPtr == NULL)
        return (0);

    writePtr = pattPtr;
    for (row = 0; row < numRows; ++row)
    {
        for (chn = 0; chn < song.antChn; ++chn)
        {
            bytes[0] = *pattPtr++;
            bytes[1] = *pattPtr++;
            bytes[2] = *pattPtr++;
            bytes[3] = *pattPtr++;
            bytes[4] = *pattPtr++;

            firstBytePtr = writePtr++;

            packBits = 0;
            if (bytes[0] > 0) { packBits |= 1; *writePtr++ = bytes[0]; } /* note */
            if (bytes[1] > 0) { packBits |= 2; *writePtr++ = bytes[1]; } /* instrument */
            if (bytes[2] > 0) { packBits |= 4; *writePtr++ = bytes[2]; } /* volume column */
            if (bytes[3] > 0) { packBits |= 8; *writePtr++ = bytes[3]; } /* effect */

            if (packBits == 15) /* first four bits set? */
            {
                /* no packing needed, write pattern data as is */

                /* point to first byte (and overwrite data) */
                writePtr = firstBytePtr;

                *writePtr++ = bytes[0];
                *writePtr++ = bytes[1];
                *writePtr++ = bytes[2];
                *writePtr++ = bytes[3];
                *writePtr++ = bytes[4];

                totalPackLen += 5;
                continue;
            }

            if (bytes[4] > 0) { packBits |= 16; *writePtr++ = bytes[4]; } /* effect parameter */

            *firstBytePtr = packBits | 128; /* write pack bits byte */
            totalPackLen += (uint16_t)(writePtr - firstBytePtr); /* bytes writen */
        }

        /* skip unused channels */
        pattPtr += (sizeof (tonTyp) * (MAX_VOICES - song.antChn));
    }

    return (totalPackLen);
}
