/* for finding memory leaks in debug mode with Visual Studio */
#if defined _DEBUG && defined _MSC_VER
#include <crtdbg.h>
#endif

#include <stdio.h>
#include <stdint.h>
#include "ft2_header.h"
#include "ft2_gui.h"
#include "ft2_pattern_ed.h"
#include "ft2_diskop.h"
#include "ft2_scopes.h"
#include "ft2_config.h"
#include "ft2_mouse.h"
#include "ft2_sample_ed.h"
#include "ft2_inst_ed.h"
#include "ft2_audio.h"
#include "ft2_wav_renderer.h"

enum
{
    WAV_FORMAT_PCM        = 0x0001,
    WAV_FORMAT_IEEE_FLOAT = 0x0003
};

typedef struct wavHeader_t
{
    uint32_t chunkID, chunkSize, format, subchunk1ID, subchunk1Size;
    uint16_t audioFormat, numChannels;
    uint32_t sampleRate, byteRate;
    uint16_t blockAlign, bitsPerSample;
    uint32_t subchunk2ID, subchunk2Size;
} wavHeader_t;

static char WAV_SysReqText[192];
static uint8_t renderBitDepth = 16, songStart, songEnd, *wavRenderBuffer;
static int16_t renderAmp;
static uint32_t renderFreq = 44100;
static SDL_Thread *thread;

void updateWavRenderer(void)
{
    uint8_t amp;
    uint32_t freq;

    fillRect(209, 116, 41, 51, PAL_DESKTOP);

    freq = renderFreq;
    charOutFast(209 + (5 * 7), 116, PAL_FORGRND, '0' + (freq % 10)); freq /= 10;
    charOutFast(209 + (4 * 7), 116, PAL_FORGRND, '0' + (freq % 10)); freq /= 10;
    charOutFast(209 + (3 * 7), 116, PAL_FORGRND, '0' + (freq % 10)); freq /= 10;
    charOutFast(209 + (2 * 7), 116, PAL_FORGRND, '0' + (freq % 10)); freq /= 10;
    charOutFast(209 + (1 * 7), 116, PAL_FORGRND, '0' + (freq % 10)); freq /= 10;
    charOutFast(209 + (0 * 7), 116, PAL_FORGRND, '0' + (freq % 10));

    amp = (uint8_t)(renderAmp);
    charOutFast(237 + (0 * 7), 130, PAL_FORGRND, '0' + (amp / 10));
    charOutFast(237 + (1 * 7), 130, PAL_FORGRND, '0' + (amp % 10));

    hexOut(237, 144, PAL_FORGRND, songStart, 2);
    hexOut(237, 158, PAL_FORGRND, songEnd,   2);
}

void updateWavRendererSettings(void) /* called when changing config.boostLevel */
{
    renderAmp = config.boostLevel;
}

void drawWavRenderer(void)
{
    drawFramework(0,   92, 291, 17, FRAMEWORK_TYPE1);
    drawFramework(0,  109,  79, 64, FRAMEWORK_TYPE1);
    drawFramework(79, 109, 212, 64, FRAMEWORK_TYPE1);

    textOutShadow(4,   96, PAL_FORGRND, PAL_DSKTOP2, "Harddisk recording:");
    textOutShadow(156, 96, PAL_FORGRND, PAL_DSKTOP2, "16-bit");
    textOutShadow(221, 96, PAL_FORGRND, PAL_DSKTOP2, "24-bit float");

    textOutShadow(85, 116, PAL_FORGRND, PAL_DSKTOP2, "Frequency");
    textOutShadow(85, 130, PAL_FORGRND, PAL_DSKTOP2, "Amplification");
    textOutShadow(85, 144, PAL_FORGRND, PAL_DSKTOP2, "Start song position");
    textOutShadow(85, 158, PAL_FORGRND, PAL_DSKTOP2, "Stop song position");

    showPushButton(PB_WAV_RENDER);
    showPushButton(PB_WAV_EXIT);
    showPushButton(PB_WAV_FREQ_UP);
    showPushButton(PB_WAV_FREQ_DOWN);
    showPushButton(PB_WAV_AMP_UP);
    showPushButton(PB_WAV_AMP_DOWN);
    showPushButton(PB_WAV_START_UP);
    showPushButton(PB_WAV_START_DOWN);
    showPushButton(PB_WAV_END_UP);
    showPushButton(PB_WAV_END_DOWN);

    /* bitdepth radiobuttons */

    if (renderBitDepth == 16)
    {
        radioButtons[RB_WAV_RENDER_BITDEPTH16].state = RADIOBUTTON_CHECKED;
        radioButtons[RB_WAV_RENDER_BITDEPTH32].state = RADIOBUTTON_UNCHECKED;
    }
    else
    {
        radioButtons[RB_WAV_RENDER_BITDEPTH16].state = RADIOBUTTON_UNCHECKED;
        radioButtons[RB_WAV_RENDER_BITDEPTH32].state = RADIOBUTTON_CHECKED;
    }

    showRadioButtonGroup(RB_GROUP_WAV_RENDER_BITDEPTH);

    updateWavRenderer();
}

void showWavRenderer(void)
{
    if (editor.ui.extended)
        exitPatternEditorExtended();

    hideTopScreen();
    showTopScreen(false);

    editor.ui.wavRendererShown = true;
    editor.ui.scopesShown = false;
    drawWavRenderer();
}

void hideWavRenderer(void)
{
    editor.ui.wavRendererShown = false;

    hidePushButton(PB_WAV_RENDER);
    hidePushButton(PB_WAV_EXIT);
    hidePushButton(PB_WAV_FREQ_UP);
    hidePushButton(PB_WAV_FREQ_DOWN);
    hidePushButton(PB_WAV_AMP_UP);
    hidePushButton(PB_WAV_AMP_DOWN);
    hidePushButton(PB_WAV_START_UP);
    hidePushButton(PB_WAV_START_DOWN);
    hidePushButton(PB_WAV_END_UP);
    hidePushButton(PB_WAV_END_DOWN);
    hideRadioButtonGroup(RB_GROUP_WAV_RENDER_BITDEPTH);

    editor.ui.scopesShown = true;
    drawScopeFramework();
}

void exitWavRenderer(void)
{
    hideWavRenderer();
}

void resetWavRenderer(void)
{
    songStart = 0;
    songEnd   = (uint8_t)(song.len - 1);

    if (editor.ui.wavRendererShown)
        updateWavRenderer();
}

static uint8_t dump_Init(uint32_t frq, int16_t amp, int16_t songPos)
{
    uint8_t i, oldMuteFlags[MAX_VOICES];

    /* *2 for stereo */
    wavRenderBuffer = (uint8_t *)(malloc(TICKS_PER_RENDER_CHUNK * ((MAX_WAV_SAMPLES_PER_TICK * 2) * (renderBitDepth / 8))));
    if (wavRenderBuffer == NULL)
        return (false);

    editor.wavIsRendering = true;

    stopPlaybackTimer();
    setPos(songPos, 0);
    playMode = PLAYMODE_SONG;
    songPlaying = true;

    /* store mute states */
    for (i = 0; i < MAX_VOICES; ++i)
        oldMuteFlags[i] = stm[i].stOff;

    resetChannels();

    /* restore mute states */
    for (i = 0; i < MAX_VOICES; ++i)
        stm[i].stOff = oldMuteFlags[i];

    setNewAudioFreq(frq);
    setAudioAmp(amp, config.masterVol, renderBitDepth == 32);

    stopVoices();
    song.globVol = 64;
    setSpeed(song.speed);

    editor.playbackSeconds = 0;
    editor.updatePlaybackTime = true;

    return (true);
}

static void dump_Close(FILE *f, uint32_t totalSamples)
{
    uint32_t tmpLen, progNameLen, songNameLen, totalBytes;
    wavHeader_t wavHeader;

    if (wavRenderBuffer != NULL)
    {
        free(wavRenderBuffer);
        wavRenderBuffer = NULL;
    }

    if (renderBitDepth == 16)
        totalBytes = totalSamples * sizeof (int16_t);
    else
        totalBytes = totalSamples * sizeof (float);

    if (totalBytes & 1)
        fputc(0, f); /* write pad byte */

    /* write LIST->INFO->INAM chunk */
    songNameLen = 0;
    while (songNameLen < 21)
    {
        if (song.name[songNameLen] == '\0')
            break;

        songNameLen++;
    }

    progNameLen = sizeof (PROG_NAME_STR) - 1;

    tmpLen = 4 + (4 + 4) + (progNameLen + 1 + ((progNameLen + 1) & 1));
    if (songNameLen > 0)
        tmpLen += ((4 + 4) + (songNameLen + 1 + ((songNameLen + 1) & 1)));

    fwrite("LIST",  sizeof (int32_t), 1, f);
    fwrite(&tmpLen, sizeof (int32_t), 1, f);
    fwrite("INFO",  sizeof (int32_t), 1, f);

    if (songNameLen > 0)
    {
        tmpLen = songNameLen + 1;
        fwrite("INAM",  sizeof (int32_t), 1, f);
        fwrite(&tmpLen, sizeof (int32_t), 1, f);
        fwrite(song.name, 1, songNameLen, f);
        fputc(0, f); /* null termination */
        if (tmpLen & 1)
            fputc(0, f); /* pad byte */
    }

    tmpLen = progNameLen + 1;
    fwrite("ISFT",  sizeof (int32_t), 1, f);
    fwrite(&tmpLen, sizeof (int32_t), 1, f);
    fwrite(PROG_NAME_STR, 1, progNameLen, f);
    fputc(0, f); /* null termination */
    if (tmpLen & 1)
        fputc(0, f); /* pad byte */

    tmpLen = ftell(f) - 8;

    /* go back and fill in WAV header */
    fseek(f, 0, SEEK_SET);

    wavHeader.chunkID       = 0x46464952; /* "RIFF" */
    wavHeader.chunkSize     = tmpLen;
    wavHeader.format        = 0x45564157; /* "WAVE" */
    wavHeader.subchunk1ID   = 0x20746D66; /* "fmt " */
    wavHeader.subchunk1Size = 16;

    if (renderBitDepth == 16)
        wavHeader.audioFormat = WAV_FORMAT_PCM;
    else
        wavHeader.audioFormat = WAV_FORMAT_IEEE_FLOAT;

    wavHeader.numChannels   = 2;
    wavHeader.sampleRate    = renderFreq;
    wavHeader.byteRate      = (wavHeader.sampleRate * wavHeader.numChannels * renderBitDepth) / 8;
    wavHeader.blockAlign    = (wavHeader.numChannels * renderBitDepth) / 8;
    wavHeader.bitsPerSample = renderBitDepth;
    wavHeader.subchunk2ID   = 0x61746164; /* "data" */
    wavHeader.subchunk2Size = totalBytes;

    /* write main header */
    fwrite(&wavHeader, 1, sizeof (wavHeader_t), f);
    fclose(f);

    stopPlaying();

    /* kludge: set speed to 6 if speed was set to 0 */
    if (song.tempo == 0)
        song.tempo = 6;

    setBackOldAudioFreq();
    setSpeed(song.speed);
    setAudioAmp(config.boostLevel, config.masterVol, config.specialFlags & BITDEPTH_24);
    editor.wavIsRendering = false;

    setMouseBusy(false);
}

static uint8_t dump_EndOfTune(int16_t endSongPos) /* exactly the same as in real FT2 */
{
    uint8_t returnValue = (editor.wavReachedEndFlag && (song.pattPos == 0) && (song.timer == 1)) || (song.tempo == 0);

    if ((song.songPos == endSongPos) && (song.pattPos == 0) && (song.timer == 1))
        editor.wavReachedEndFlag = true;

    return (returnValue);
}

uint32_t dump_RenderTick(uint8_t *buffer) /* returns bytes mixed */
{
    replayerBusy = true;

    mix_SaveIPVolumes();
    mainPlayer();
    mix_UpdateChannelVolPanFrq();

    replayerBusy = false;

    return (mixReplayerTickToBuffer(buffer, renderBitDepth));
}

int32_t SDLCALL renderWavThread(void *ptr)
{
    uint8_t *ptr8, renderDone;
    uint32_t i, samplesInChunk, tickSamples, sampleCounter;
    double dSecondsElapsed;
    FILE *f;

    f = (FILE *)(editor.wavRendererFileHandle);
    fseek(f, sizeof (wavHeader_t), SEEK_SET);

    pauseAudio();

    editor.wavSongStart = (uint8_t)(MAX(0, MIN(songStart, song.len - 1)));
    editor.wavSongEnd   = (uint8_t)(MAX(0, MIN(MAX(songStart, songEnd), song.len - 1)));

    if (!dump_Init(renderFreq, renderAmp, editor.wavSongStart))
    {
        resumeAudio();
        setMouseBusy(false);
        sysReqQueue(SR_OOM_ERROR); /* out of memory */
        return (true);
    }

    sampleCounter = 0;
    renderDone = false;

    editor.wavReachedEndFlag = false;
    while (!renderDone && editor.wavIsRendering)
    {
        samplesInChunk = 0;
        ptr8 = wavRenderBuffer;

        /* render several ticks at once to prevent frequent disk I/O (speeds up the process) */
        for (i = 0; i < TICKS_PER_RENDER_CHUNK; ++i)
        {
            if (!editor.wavIsRendering || dump_EndOfTune(editor.wavSongEnd))
            {
                renderDone = true;
                break;
            }

            tickSamples = dump_RenderTick(ptr8) * 2; /* *2 for stereo */

            samplesInChunk += tickSamples;
            sampleCounter  += tickSamples;

            /* increase buffer pointer */
            if (renderBitDepth == 16)
                ptr8 += (tickSamples * sizeof (int16_t));
            else
                ptr8 += (tickSamples * sizeof (float));

            /* update playback time and pattern editor positions */

            dSecondsElapsed = (sampleCounter / 2.0) / renderFreq;
            double2int32_round(editor.playbackSeconds, dSecondsElapsed);

            editor.editPattern = (uint8_t)(song.pattNr);
            editor.songPos = song.songPos;

            editor.updatePatternEditor = true;
            editor.updatePosSections   = true;

            editor.updatePlaybackTime  = true;
        }

        /* write buffer to disk */
        if (samplesInChunk > 0)
        {
            if (renderBitDepth == 16)
                fwrite(wavRenderBuffer, sizeof (int16_t), samplesInChunk, f);
            else
                fwrite(wavRenderBuffer, sizeof (float), samplesInChunk, f);
        }
    }

    /* update playback time and pattern editor positions (again) */

    dSecondsElapsed = (sampleCounter / 2.0) / renderFreq;
    double2int32_round(editor.playbackSeconds, dSecondsElapsed);

    editor.editPattern = (uint8_t)(song.pattNr);
    editor.songPos = song.songPos;

    editor.updatePatternEditor = true;
    editor.updatePosSections   = true;
    editor.updatePlaybackTime  = true;

    dump_Close(f, sampleCounter);

    resumeAudio();

    editor.diskOpReadOnOpen = true;

    (void)(ptr); /* prevent compiler warning */
    return (true);
}

static void createOverwriteText(char *name)
{
    char nameTmp[128];
    uint32_t nameLen;

    /* read entry name to a small buffer */
    nameLen = (uint32_t)(strlen(name));
    memcpy(nameTmp, name, (nameLen >= sizeof (nameTmp)) ? sizeof (nameTmp) : (nameLen + 1));
    nameTmp[sizeof (nameTmp) - 1] = '\0';

    trimEntryName(nameTmp, false);

    sprintf(WAV_SysReqText, "Overwrite file \"%s\"?", nameTmp);
    sysReqs[SR_WAV_OVERWRITE].text = WAV_SysReqText;
}

void wavRender(uint8_t checkOverwrite)
{
    char *filename;

    hideSystemRequest();

    if (songStart > (song.len - 1))
        songStart = (uint8_t)(song.len - 1);

    if (songEnd > (song.len - 1))
        songEnd = (uint8_t)(song.len - 1);

    if (songStart > songEnd)
        songStart = songEnd;

    updateWavRenderer();
    diskOpChangeFilenameExt(".wav");

    filename = getDiskOpFilename();
    if (checkOverwrite && fileExistsAnsi(filename))
    {
        createOverwriteText(filename);
        sysReqQueue(SR_WAV_OVERWRITE);
        return;
    }

    editor.wavRendererFileHandle = fopen(filename, "wb");
    if (editor.wavRendererFileHandle == NULL)
    {
        sysReqQueue(SR_WAV_WRITE_ERROR);
        return;
    }

    setMouseBusy(true);
    thread = SDL_CreateThread(renderWavThread, "FT2 Clone Wav Renderer Thread", NULL);
    if (thread == NULL)
    {
        fclose((FILE *)(editor.wavRendererFileHandle));
        setMouseBusy(false);
        sysReqQueue(SR_THREAD_ERROR);
        return;
    }

    /* don't let thread wait for this thread, let it clean up on its own when done */
    SDL_DetachThread(thread);
}

void wavRenderOverwrite(void) /* from sys req. */
{
    wavRender(false);
}

void pbWavRender(void)
{
    wavRender(config.cfg_OverwriteWarning);
}

void pbWavExit(void)
{
    exitWavRenderer();
}

void pbWavFreqUp(void)
{
    if (renderFreq < MAX_WAV_AUDIO_FREQ)
    {
        if (renderFreq == 8000)
            renderFreq = 11025;
        else if (renderFreq == 11025)
            renderFreq = 16000;
        else if (renderFreq == 16000)
            renderFreq = 22050;
        else if (renderFreq == 22050)
            renderFreq = 32000;
        else if (renderFreq == 32000)
            renderFreq = 44100;
        else if (renderFreq == 44100)
            renderFreq = 48000;
        else if (renderFreq == 48000)
            renderFreq = 96000;
        else if (renderFreq == 96000)
            renderFreq = 192000;

        updateWavRenderer();
    }
}

void pbWavFreqDown(void)
{
    if (renderFreq > MIN_WAV_AUDIO_FREQ)
    {
        if (renderFreq == 192000)
            renderFreq = 96000;
        else if (renderFreq == 96000)
            renderFreq = 48000;
        else if (renderFreq == 48000)
            renderFreq = 44100;
        else if (renderFreq == 44100)
            renderFreq = 32000;
        else if (renderFreq == 32000)
            renderFreq = 22050;
        else if (renderFreq == 22050)
            renderFreq = 16000;
        else if (renderFreq == 16000)
            renderFreq = 11025;
        else if (renderFreq == 11025)
            renderFreq = 8000;

        updateWavRenderer();
    }
}

void pbWavAmpUp(void)
{
    if (renderAmp < 32)
    {
        renderAmp++;
        updateWavRenderer();
    }
}

void pbWavAmpDown(void)
{
    if (renderAmp > 1)
    {
        renderAmp--;
        updateWavRenderer();
    }
}

void pbWavSongStartUp(void)
{
    if (songStart < (song.len - 1))
    {
        if (++songEnd < songStart)
              songEnd = songStart;

        updateWavRenderer();
    }
}

void pbWavSongStartDown(void)
{
    if (songStart > 0)
    {
        songStart--;
        updateWavRenderer();
    }
}

void pbWavSongEndUp(void)
{
    if (songEnd < (song.len - 1))
    {
        songEnd++;
        updateWavRenderer();
    }
}

void pbWavSongEndDown(void)
{
    if (songEnd > songStart)
    {
        songEnd--;
        updateWavRenderer();
    }
}

void rbWavRenderBitDepth16(void)
{
    checkRadioButton(RB_WAV_RENDER_BITDEPTH16);
    renderBitDepth = 16;
}

void rbWavRenderBitDepth32(void)
{
    checkRadioButton(RB_WAV_RENDER_BITDEPTH32);
    renderBitDepth = 32;
}
